@@ -29,6 +29,9 @@ class sqlApp:
         c = self.config = get_config(odbclirc_file)
         self.initialize_logging()
         self.set_default_pager(c)
+        self.mouse_support: bool = c["main"].as_bool("mouse_support")
+        self.preview_limit_rows = c["main"].as_int("preview_limit_rows")
+        self.preview_chunk_size = c["main"].as_int("preview_fetch_chunk_size")
         self.pager_reserve_lines = c["main"].as_int("pager_reserve_lines")
         self.table_format = c["main"]["table_format"]
         self.timing_enabled = c["main"].as_bool("timing")
@@ -46,7 +49,7 @@ class sqlApp:
         self.show_login_prompt: bool = False
         self.show_preview: bool = False
         self.show_disconnect_dialog: bool = False
-        self.active_conn = None
+        self._active_conn = None
         self.obj_list = []
         dsns = list(datasources().keys())
         if len(dsns) < 1:
@@ -62,10 +65,18 @@ class sqlApp:
         # Loop over side-bar when moving past the element on the bottom
         self.obj_list[len(self.obj_list) - 1].next_object = self.obj_list[0]
         self._selected_object = self.obj_list[0]
-        self.completer = MssqlCompleter(smart_completion=True, my_app = self)
+        self.completer = MssqlCompleter(smart_completion = True, get_conn = lambda: self.active_conn)
 
         self.application = self._create_application()
 
+    @property
+    def active_conn(self) -> sqlConnection:
+        return self._active_conn
+
+    @active_conn.setter
+    def active_conn(self, conn: sqlConnection) -> None:
+        self._active_conn = conn
+
     @property
     def selected_object(self) -> myDBObject:
         return self._selected_object
@@ -241,9 +252,8 @@ class sqlApp:
                 event.app.layout.focus(self.sql_layout.lprompt)
             if type(obj).__name__ == "myDBConn" and obj.conn.connected():
                 # OG: some thread locking may be needed here
-                self.completer.reset_completions()
-                self.active_conn = obj.conn
-            elif type(obj).__name__ == "myDBTable":
+                self._active_conn = obj.conn
+            elif obj.otype in ["table", "view", "function"]:
                 self.show_preview = True
                 self.show_sidebar = False
                 event.app.layout.focus(self.sql_layout.preview)
@@ -278,7 +288,7 @@ class sqlApp:
             enable_page_navigation_bindings = True,
             style = style_factory(self.syntax_style, self.cli_style),
             include_default_pygments_style = False,
-            mouse_support = True,
+            mouse_support = self.mouse_support,
             full_screen = False,
             editing_mode = EditingMode.VI if self.config["main"].as_bool("vi") else EditingMode.EMACS
         )
