@@ -7,12 +7,14 @@
 from __future__ import print_function, unicode_literals
 import logging
 import re
+from typing import Callable
 from itertools import count, chain
 import operator
 from collections import namedtuple, defaultdict, OrderedDict
 from prompt_toolkit.completion import Completer, Completion, PathCompleter
 from prompt_toolkit.document import Document
-from .sqlcompletion import (FromClauseItem, suggest_type, Special, NamedQuery,
+from ..conn import sqlConnection
+from .sqlcompletion import (Blank, FromClauseItem, suggest_type, Special, NamedQuery,
                                              Database, Schema, Table, Function, Column, View,
                                              Keyword, Datatype, Alias, Path, JoinCondition, Join)
 from .parseutils.meta import ColumnMetadata, ForeignKey
@@ -20,7 +22,6 @@ from .parseutils.utils import last_word
 from .parseutils.tables import TableReference
 from .mssqlliterals.main import get_literals
 from .prioritization import PrevalenceCounter
-# from mssqlcli.util import decode
 
 Match = namedtuple('Match', ['completion', 'priority'])
 
@@ -75,12 +76,12 @@ class MssqlCompleter(Completer):
 
     def __init__(
             self,
-            my_app: "sqlApp",
+            get_conn: Callable[[], sqlConnection],
             smart_completion=True,
             settings=None):
         super(MssqlCompleter, self).__init__()
         self.smart_completion = smart_completion
-        self.my_app = my_app
+        self._get_conn = get_conn
         self.prioritizer = PrevalenceCounter()
         settings = settings or {}
         self.signature_arg_style = settings.get(
@@ -115,8 +116,6 @@ class MssqlCompleter(Completer):
         self.name_pattern = re.compile(r"^[_a-z][_a-z0-9\$]*$")
 
         self.databases = []
-        self.dbmetadata = {'table': {}, 'view': {}, 'functions': {},
-                           'datatypes': {}}
         self.search_path = []
         self.casing = {}
 
@@ -129,12 +128,13 @@ class MssqlCompleter(Completer):
         self.logger = logging.getLogger(__name__)
         self.logger.debug("Completer instantiated")
 
+    @property
+    def active_conn(self) -> sqlConnection:
+        return self._get_conn()
+
     def escape_name(self, name):
-        # OG: double quotation here, probably needs to be something that
-        # can be configured
-        if name and self.my_app.active_conn is not None:
-            qtchar = self.my_app.active_conn.quotechar
-            name = (qtchar + "%s" + qtchar) % name
+        if self.active_conn is not None:
+            name = self.active_conn.escape_name(name)
 
         return name
 
@@ -143,18 +143,19 @@ class MssqlCompleter(Completer):
 
     def unescape_name(self, name):
         """ Unquote a string."""
-        if self.my_app.active_conn is not None:
-            qtchar = self.my_app.active_conn.quotechar
-            if name and name[0] == qtchar and name[-1] == qtchar:
-                name = name[1:-1]
+        if self.active_conn is not None:
+            name = self.active_conn.unescape_name(name)
 
         return name
 
-    def escaped_names(self, names):
-        return [self.escape_name(name) for name in names]
+    def escape_names(self, names):
+        if self.active_conn is not None:
+            names = self.active_conn.escape_names(names)
+
+        return names
 
     def extend_database_names(self, databases):
-        databases = self.escaped_names(databases)
+        databases = self.escape_names(databases)
         self.databases.extend(databases)
 
     def extend_keywords(self, additional_keywords):
@@ -162,23 +163,6 @@ class MssqlCompleter(Completer):
         # OG: Unclear what the roll of all_completions is
         # self.all_completions.update(additional_keywords)
 
-    def extend_schemas(self, schemas):
-
-        # schemas is a list of schema names
-        schemas = self.escaped_names(schemas)
-        # OG: seems like this is done in for loop below, no?
-        # metadata = self.dbmetadata['tables']
-        #for schema in schemas:
-        #    metadata[schema] = {}
-
-        # dbmetadata.values() are the 'tables' and 'functions' dicts
-        for metadata in self.dbmetadata.values():
-            for schema in schemas:
-                metadata[schema] = {}
-
-        # OG: Unclear what the roll of all_completions is
-        # self.all_completions.update(schemas)
-
     def extend_casing(self, words):
         """ extend casing data
 
@@ -187,70 +171,23 @@ class MssqlCompleter(Completer):
         # casing should be a dict {lowercasename:PreferredCasingName}
         self.casing = dict((word.lower(), word) for word in words)
 
-    def extend_relations(self, data, kind):
-        """extend metadata for tables or views.
-
-        :param data: list of (catalog_name, schema_name, rel_name) tuples
-        :param kind: either 'tables' or 'views'
-
-        :return:
-
-        """
-
-        data = [self.escaped_names(d) for d in data]
-
-        # dbmetadata['tables']['schema_name']['table_name'] should be an
-        # OrderedDict {column_name:ColumnMetaData}.
-        metadata = self.dbmetadata[kind]
-        for catalog, schema, relname in data:
-            try:
-                metadata[catalog][schema][relname] = OrderedDict()
-            except KeyError:
-                self.logger.error('%r %r listed in unrecognized schema %r',
-                              kind, relname, schema)
-
-            # OG: Unclear what the roll of all_completions is
-            # self.all_completions.add(relname)
-
-    def extend_columns(self, column_data, kind):
-        """extend column metadata.
-
-        :param column_data: list of (catalog_name, schema_name, rel_name, column_name,
-        column_type, has_default, default) tuples
-        :param kind: either 'tables' or 'views'
-
-        :return:
-
-        """
-        metadata = self.dbmetadata[kind]
-        for catalog, schema, relname, colname, datatype, default in column_data:
-            (catalog, schema, relname, colname) = self.escaped_names(
-                [catalog, schema, relname, colname])
-            column = ColumnMetadata(
-                name=colname,
-                datatype=datatype,
-                has_default=default,
-                default=default
-            )
-            metadata[catalog][schema][relname][colname] = column
-            # OG: Unclear what the roll of all_completions is
-            # self.all_completions.add(colname)
-
     def extend_functions(self, func_data):
 
         # func_data is a list of function metadata namedtuples
 
         # dbmetadata['schema_name']['functions']['function_name'] should return
         # the function metadata namedtuple for the corresponding function
-        metadata = self.dbmetadata['functions']
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
+        submeta = metadata['function']
 
         for f in func_data:
-            schema, func = self.escaped_names([f.schema_name, f.func_name])
+            schema, func = self.escape_names([f.schema_name, f.func_name])
 
-            if func in metadata[schema]:
-                metadata[schema][func].append(f)
+            if func in submeta[schema]:
+                submeta[schema][func].append(f)
             else:
-                metadata[schema][func] = [f]
+                submeta[schema][func] = [f]
 
             # OG: Unclear what the roll of all_completions is
             # self.all_completions.add(func)
@@ -262,10 +199,12 @@ class MssqlCompleter(Completer):
         # This is used when suggesting functions, to avoid the latency that would result
         # if we'd recalculate the arg lists each time we suggest functions (in
         # large DBs)
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
         self._arg_list_cache = {
             usage: {
                 meta: self._arg_list(meta, usage)
-                for sch, funcs in self.dbmetadata['functions'].items()
+                for sch, funcs in metadata['function'].items()
                 for func, metas in funcs.items()
                 for meta in metas
             }
@@ -280,15 +219,18 @@ class MssqlCompleter(Completer):
 
         # These are added as a list of ForeignKey namedtuples to the
         # ColumnMetadata namedtuple for both the child and parent
-        meta = self.dbmetadata['table']
+        # OG: This needs catalog facelift
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
+        submeta = metadata['table']
 
         for fk in fk_data:
-            e = self.escaped_names
+            e = self.escape_names
             parentschema, childschema = e([fk.parentschema, fk.childschema])
             parenttable, childtable = e([fk.parenttable, fk.childtable])
             childcol, parcol = e([fk.childcolumn, fk.parentcolumn])
-            childcolmeta = meta[childschema][childtable][childcol]
-            parcolmeta = meta[parentschema][parenttable][parcol]
+            childcolmeta = submeta[childschema][childtable][childcol]
+            parcolmeta = submeta[parentschema][parenttable][parcol]
             fk = ForeignKey(parentschema, parenttable, parcol,
                             childschema, childtable, childcol)
             childcolmeta.foreignkeys.append((fk))
@@ -299,11 +241,12 @@ class MssqlCompleter(Completer):
         # dbmetadata['datatypes'][schema_name][type_name] should store type
         # metadata, such as composite type field names. Currently, we're not
         # storing any metadata beyond typename, so just store None
-        meta = self.dbmetadata['datatypes']
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
 
         for t in type_data:
-            schema, type_name = self.escaped_names(t)
-            meta[schema][type_name] = None
+            schema, type_name = self.escape_names(t)
+            metadata["datatype"][schema][type_name] = None
             # OG: Unclear what the roll of all_completions is
             # self.all_completions.add(type_name)
 
@@ -316,14 +259,17 @@ class MssqlCompleter(Completer):
             self.prioritizer.update(text)
 
     def set_search_path(self, search_path):
-        self.search_path = self.escaped_names(search_path)
+        self.search_path = self.escape_names(search_path)
 
     def reset_completions(self):
+        # databases at this point is not used
         self.databases = []
+        # special_commands at this point is not used
         self.special_commands = []
+        # search_path at this point is not used
         self.search_path = []
-        self.dbmetadata = {'table': {}, 'view': {}, 'functions': {},
-                           'datatypes': {}} 
+        conn = self.active_conn
+        conn.dbmetadata.reset_metadata()
         # OG: Unclear what the roll of all_completions is
         #self.all_completions = set(self.keywords + self.functions)
 
@@ -412,6 +358,9 @@ class MssqlCompleter(Completer):
                 syn_matches = [m for m in syn_matches if m]
                 sort_key = max(syn_matches) if syn_matches else None
             else:
+                if cand == "":
+                    # We don't offer up empty strings as suggestions
+                    continue
                 item, display_meta, prio, prio2, display = cand, meta, 0, 0, cand
                 sort_key = _match(cand)
 
@@ -681,6 +630,10 @@ class MssqlCompleter(Completer):
     # TODO: Need to account for suggestion.catalog
     def get_function_matches(
             self, suggestion, word_before_cursor, alias=False):
+
+        # OG: hack early exit, for now
+        return []
+        # We'll have to do away with this right? How can we possibly know this
         if suggestion.usage == 'from':
             # Only suggest functions allowed in FROM clause
             def filt(f):
@@ -703,6 +656,7 @@ class MssqlCompleter(Completer):
 
         matches = self.find_matches(word_before_cursor, funcs, meta='function')
 
+        # OG: TODO: catalog here
         if not suggestion.schema and not suggestion.usage:
             # also suggest hardcoded functions using startswith matching
             predefined_funcs = self.find_matches(
@@ -713,19 +667,26 @@ class MssqlCompleter(Completer):
         return matches
 
     def get_schema_matches(self, suggestion, word_before_cursor):
-        conn = self.my_app.active_conn
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
+        submeta = metadata['table']
         if suggestion.parent:
             catalog_u = self.unescape_name(suggestion.parent)
         else:
             catalog_u = conn.current_catalog()
 
         catalog_e = self.escape_name(catalog_u)
-        cats = self.dbmetadata['table'].keys()
+        self.logger.debug("get_schema_matches: parent %s", suggestion.parent)
         # OG: Note here, if there is even a single schema in [catalog_e].keys()
         # we'll happily return a potentially incomplete result set.
-        if catalog_e in cats and self.dbmetadata['table'][catalog_e]:
-            schema_names_e = self.dbmetadata['table'][catalog_e].keys()
-        else: 
+        schema_names_e = conn.dbmetadata.get_schemas(catalog = catalog_e)
+
+        if schema_names_e is None:
+            # Asking for schema in a non-existant catalog
+            return []
+
+        if len(schema_names_e) == 0:
+            # Catalog exists in dbmetadata but is empty
             if suggestion.parent:
                 # Looking for schemas in a specified catalog
                 schema_names = []
@@ -739,24 +700,22 @@ class MssqlCompleter(Completer):
                             schema = "",
                             table = "",
                             type = "")
-                    for r in res:
-                        if (r.schema not in schema_names and r.schema != ""):
-                            schema_names.append(r.schema)
+                    schema_names = [r.schema for r in res]
             else:
                 # Looking for schemas in current catalog
-                schema_names = set(conn.list_schemas())
+                schema_names = conn.list_schemas()
+
+            schema_names = set(schema_names)
             
-            schema_names_e = self.escaped_names(schema_names)
-            if len(schema_names_e):
-                for metadata in self.dbmetadata.values():
-                    metadata[catalog_e] = {}
-                    for schema_e in schema_names_e:
-                        self.logger.debug("get_schema_matches: Creating dict %s.%s", catalog_e, schema_e)
-                        metadata[catalog_e][schema_e] = {}
+            schema_names_e = self.escape_names(schema_names)
+            conn.dbmetadata.extend_schemas(catalog = catalog_e, names = schema_names_e)
 
         return self.find_matches(
             word_before_cursor, schema_names_e, meta='schema')
 
+    def get_blank_item_matches(self, suggestion, word_before_cursor):
+        return []
+
     def get_from_clause_item_matches(self, suggestion, word_before_cursor):
         alias = self.generate_aliases
         s = suggestion
@@ -870,14 +829,14 @@ class MssqlCompleter(Completer):
                                  meta='table alias')
 
     def get_database_matches(self, _, word_before_cursor):
-        catalogs = self.dbmetadata["table"].keys()
-        if not catalogs and (self.my_app.active_conn.connected()):
-            catalogs = self.escaped_names(
-                    self.my_app.active_conn.list_catalogs())
-            for reltype in ('table', 'view'):
-                self.logger.debug("get_database_matches: Populating catalogs %s", reltype)
-                self.dbmetadata[reltype].update(dict.fromkeys(catalogs, {}))
-        return self.find_matches(word_before_cursor, catalogs,
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
+        catalogs_e = conn.dbmetadata.get_catalogs()
+        if catalogs_e is None and (conn.connected()):
+            catalogs_e = self.escape_names(conn.list_catalogs())
+            conn.dbmetadata.extend_catalogs(catalogs_e)
+
+        return self.find_matches(word_before_cursor, catalogs_e,
                                  meta='catalog')
 
     def get_keyword_matches(self, suggestion, word_before_cursor):
@@ -937,6 +896,7 @@ class MssqlCompleter(Completer):
 #            word_before_cursor, named_queries.list(), meta='named query')
 
     suggestion_matchers = {
+        Blank: get_blank_item_matches,
         FromClauseItem: get_from_clause_item_matches,
         JoinCondition: get_join_condition_matches,
         Join: get_join_matches,
@@ -962,9 +922,10 @@ class MssqlCompleter(Completer):
         :return: {TableReference:{colname:ColumnMetaData}}
 
         """
+        conn = self.active_conn
         ctes = dict((normalize_ref(t.name), t.columns) for t in local_tbls)
         columns = OrderedDict()
-        meta = self.dbmetadata
+        metadata = conn.dbmetadata.data
 
         def addcols(schema, rel, alias, reltype, cols):
             tbl = TableReference(schema, rel, alias, reltype == 'functions')
@@ -985,14 +946,14 @@ class MssqlCompleter(Completer):
                 if tbl.is_function:
                     # Return column names from a set-returning function
                     # Get an array of FunctionMetadata objects
-                    functions = meta['functions'].get(schema, {}).get(relname)
+                    functions = metadata['function'].get(schema, {}).get(relname)
                     for func in (functions or []):
                         # func is a FunctionMetadata object
                         cols = func.fields()
                         addcols(schema, relname, tbl.alias, 'functions', cols)
                 else:
                     for reltype in ('table', 'view'):
-                        cols = meta[reltype].get(schema, {}).get(relname)
+                        cols = metadata[reltype].get(schema, {}).get(relname)
                         if cols:
                             cols = cols.values()
                             addcols(schema, relname, tbl.alias, reltype, cols)
@@ -1018,13 +979,20 @@ class MssqlCompleter(Completer):
             if tbl.catalog:
                 catalog_u = self.unescape_name(tbl.catalog)
             else:
-                catalog_u = self.my_app.active_conn.current_catalog()
+                catalog_u = self.active_conn.current_catalog()
 
             # TODO: What if no schema? Possible in some DBMS
             if tbl.schema:
                 schema_u = self.unescape_name(tbl.schema)
             else:
                 schema_u = ""
+
+            if catalog_u is None or catalog_u == "":
+                # Don't allow "".[schema].[table]
+                # Interpret this to mean [schema]."".[table]
+                catalog_u = schema_u
+                schema_u = ""
+
             relname_u = self.unescape_name(tbl.name)
             catalog = self.escape_name(catalog_u)
             schema = self.escape_name(schema_u)
@@ -1040,9 +1008,10 @@ class MssqlCompleter(Completer):
                 #    cols = func.fields()
                 #    addcols(schema, relname, tbl.alias, 'functions', cols)
             else:
-                conn = self.my_app.active_conn
+                conn = self.active_conn
+                # Per SQLColumns spec: CatalogName cannot contain a string search pattern
                 res = conn.find_columns(
-                        catalog = conn.sanitize_search_string(catalog_u),
+                        catalog = catalog_u,
                         schema = conn.sanitize_search_string(schema_u),
                         table = conn.sanitize_search_string(relname_u),
                         column = "%")
@@ -1063,11 +1032,13 @@ class MssqlCompleter(Completer):
         :param schema is the schema qualification input by the user (if any)
 
         """
-        metadata = self.dbmetadata[obj_typ]
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
+        submeta = metadata[obj_typ]
         if schema:
             schema = self.escape_name(schema)
-            return [schema] if schema in metadata else []
-        return self.search_path if self.search_path_filter else metadata.keys()
+            return [schema] if schema in submeta else []
+        return self.search_path if self.search_path_filter else submeta.keys()
 
     def _maybe_schema(self, schema, parent):
         return None if parent or schema in self.search_path else schema
@@ -1078,6 +1049,8 @@ class MssqlCompleter(Completer):
         :param schema is the schema qualification input by the user (if any)
 
         """
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
 
         return [
             SchemaObject(
@@ -1085,7 +1058,7 @@ class MssqlCompleter(Completer):
                 schema=(self._maybe_schema(schema=sch, parent=schema))
             )
             for sch in self._get_schemas(obj_type, schema)
-            for obj in self.dbmetadata[obj_type][sch].keys()
+            for obj in metadata[obj_type][sch].keys()
         ]
 
     def populate_objects(self, catalog, schema, obj_type):
@@ -1095,35 +1068,23 @@ class MssqlCompleter(Completer):
 
         """
         ret = []
-        conn = self.my_app.active_conn
+        obj_names = []
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
+        submeta = metadata[obj_type]
+        self.logger.debug("populate_objects(%s): Called for %s.%s",
+                obj_type, catalog, schema)
         if catalog is None and schema is None:
-            # Query free tables, no catalog or schema (think SQLite)
-            # Trying to communicate here empty string, "" to
-            # SQLTables API.  Unfortunately, nanodbc intercepts these
-            # and translates them into nullptr which in turn are treated
-            # as wildcards per ODBC standard.  So try to work around the
-            # nanodbc intercept, and create an empty string as a
-            # null-terminated string without content.
-            res = conn.find_tables(
-                    catalog = "\x00",
-                    schema = "\x00",
-                    table = "",
-                    type = obj_type)
-
-            for r in res:
-                name_e = self.escape_name(r.name)
-                ret.append(
-                    SchemaObject(
-                        name = name_e,
-                        schema = "",
-                        catalog = ""
-                    )
-                )
-            # Don't expand dbmetadata with free tables (for now)
-            return ret
+            catalog = ""
+            schema = ""
         if catalog is None:
-            #Set to current catalog
+            # Set to current catalog
             catalog = conn.current_catalog()
+            if catalog is None or catalog == "":
+                # Don't allow "".[schema]
+                # Interpret this to mean [schema].""
+                catalog = schema
+                schema = ""
         # Note to self, as soon as a period is inputted a schema (parent) is
         # no longer none.  Technically we should never be in a situation where
         # catalog is not None but schema is None.
@@ -1131,31 +1092,32 @@ class MssqlCompleter(Completer):
         # dbmetadata always escaped?
         catalog_e = self.escape_name(self.unescape_name(catalog))
         schema_e = self.escape_name(self.unescape_name(schema))
+        obj_names = conn.dbmetadata.get_objects(catalog = catalog_e, schema = schema_e, obj_type = obj_type)
+        if obj_names is None:
+            self.logger.debug("populate_objects(%s): Called for %s.%s, catalog/schema not found",
+                    obj_type, catalog, schema)
+            return []
 
-        metadata = self.dbmetadata[obj_type]
-        if catalog_e in metadata.keys() and \
-                schema_e in metadata[catalog_e].keys() and \
-                metadata[catalog_e][schema_e]:
-            self.logger.debug("populate_objects(%s): Found %s.%s metadata", obj_type, catalog_e, schema_e)
-            obj_names = metadata[catalog_e][schema_e].keys()
-            for name_e in obj_names:
-                ret.append(
-                    SchemaObject(
-                        name=name_e,
-                        schema=schema_e, #should this be r.schema
-                        catalog=catalog_e #should this be r.catalog
-                    )
-                )
-        else:
-            self.logger.debug("populate_objects(%s): Did not find %s.%s metadata.  Will query.", obj_type, catalog_e, schema_e)
+        if len(obj_names) == 0:
+            # catalog.schema were found but dbmetadata had no information as to
+            # content.  So let's attempt to query
             obj_names = []
-            res = conn.find_tables(
-                    catalog = conn.sanitize_search_string(
-                        self.unescape_name(catalog)),
-                    schema = conn.sanitize_search_string(
-                        self.unescape_name(schema)),
-                    table = "",
-                    type = obj_type)
+            self.logger.debug("populate_objects(%s): Did not find %s.%s metadata.  Will query.", obj_type, catalog_e, schema_e)
+            # Special case: Look for tables without catalog/schema
+            if catalog == "" and schema == "":
+                res = conn.find_tables(
+                        catalog = "\x00",
+                        schema = "\x00",
+                        table = "",
+                        type = obj_type)
+            else:
+                res = conn.find_tables(
+                        catalog = conn.sanitize_search_string(
+                            self.unescape_name(catalog)),
+                        schema = conn.sanitize_search_string(
+                            self.unescape_name(schema)),
+                        table = "",
+                        type = obj_type)
             for r in res:
                 name_e = self.escape_name(r.name)
                 ret.append(
@@ -1167,11 +1129,17 @@ class MssqlCompleter(Completer):
                 )
                 obj_names.append(name_e)
             self.logger.debug("populate_objects(%s): Query complete %s.%s", obj_type, catalog_e, schema_e)
-            if catalog_e in metadata.keys() and \
-                    schema_e in metadata[catalog_e].keys():
-                self.logger.debug("populate_objects(%s): Populating objects %s.%s", obj_type, catalog_e, schema_e)
-                metadata[catalog_e][schema_e].update(
-                        dict.fromkeys(obj_names, {})
+            conn.dbmetadata.extend_objects(
+                    catalog = catalog_e, schema = schema_e,
+                    names = obj_names, obj_type = obj_type)
+        else:
+            for name_e in obj_names:
+                ret.append(
+                    SchemaObject(
+                        name=name_e,
+                        schema=schema_e, #should this be r.schema
+                        catalog=catalog_e #should this be r.catalog
+                    )
                 )
         return ret
 
@@ -1184,6 +1152,8 @@ class MssqlCompleter(Completer):
 
         """
 
+        conn = self.active_conn
+        metadata = conn.dbmetadata.data
         # Because of multiple dispatch, we can have multiple functions
         # with the same name, which is why `for meta in metas` is necessary
         # in the comprehensions below
@@ -1193,8 +1163,8 @@ class MssqlCompleter(Completer):
                 schema=(self._maybe_schema(schema=sch, parent=schema)),
                 meta=meta
             )
-            for sch in self._get_schemas('functions', schema)
-            for (func, metas) in self.dbmetadata['functions'][sch].items()
+            for sch in self._get_schemas('function', schema)
+            for (func, metas) in metadata['function'][sch].items()
             for meta in metas
             if filter_func(meta)
         ]
