@@ -2,11 +2,12 @@ from enum import Enum
 from cyanodbc import connect, Connection, SQLGetInfo, Cursor, DatabaseError, ConnectError
 from typing import Optional
 from cli_helpers.tabular_output import TabularOutputFormatter
-from multiprocessing import Process, Pipe
 from logging import getLogger
 from re import sub
-from threading import Lock
-from .executor import executor_process, cmsg, commandStatus
+from threading import Lock, Event, Thread, Condition
+from asyncio import get_event_loop
+from enum import IntEnum
+from .dbmetadata import DbMetadata
 
 formatter = TabularOutputFormatter()
 
@@ -17,6 +18,11 @@ class connStatus(Enum):
     FETCHING = 3
     ERROR = 4
 
+class executionStatus(IntEnum):
+    OK = 0
+    FAIL = 1
+    OKWRESULTS = 2
+
 class sqlConnection:
     def __init__(
         self,
@@ -31,10 +37,8 @@ class sqlConnection:
         self.query: str = None
         self.username = username
         self.password = password
-        self.status = connStatus.DISCONNECTED
-        self.executor: Process = None
-        self.parent_chan, self.child_chan = Pipe()
         self.logger = getLogger(__name__)
+        self.dbmetadata = DbMetadata()
         self._quotechar = None
         self._search_escapechar = None
         self._search_escapepattern = None
@@ -49,6 +53,50 @@ class sqlConnection:
         # to return.
         self._lock = Lock()
 
+        # Lock to be held when updating self.status, which can happen from
+        # a thread
+        self._status_lock = Lock()
+        # Lock to be held when updating self._execution_status_lock which
+        # can happen from a thread
+        self._execution_status_lock = Lock()
+
+        # Lock that protects interaction with _fetch_res
+        self._fetch_cv = Condition()
+        # This is the list that carries the cache of retrieved rows via the
+        # asynchronous fetch operation
+        self._fetch_res: list = []
+        self._fetch_thread = Thread()
+        self._cancel_async_event = Event()
+
+        self._status = connStatus.DISCONNECTED
+        self._execution_status: executionStatus = executionStatus.OK
+        self._execution_err: str = None
+
+    @property
+    def execution_status(self) -> executionStatus:
+        """ Hold the lock here since it gets assigned in execute
+            which can be called in a different thread """
+        with self._execution_status_lock:
+            res = self._execution_status
+        return res
+
+    @property
+    def status(self) -> connStatus:
+        """ Hold the lock here since it can be assigned in more than one
+            thread """
+        with self._status_lock:
+            res = self._status
+        return res
+
+    @property
+    def execution_err(self) -> str:
+        """ Last execution error: Cleared prior to every execution.
+            Hold the lock here since it gets assigned in execute
+            which can be called in a different thread """
+        with self._lock:
+            res = self._execution_err
+        return res
+
     @property
     def quotechar(self) -> str:
         if self._quotechar is None:
@@ -70,23 +118,59 @@ class sqlConnection:
         if self._search_escapepattern is None:
             # https://stackoverflow.com/questions/2428117/casting-raw-strings-python
             self._search_escapepattern = \
-                (self.search_escapechar).encode("unicode-escape").decode() + "\\1"
+                (self.search_escapechar).encode("unicode-escape").decode()
 
         return self._search_escapepattern
 
+    def update_status(self, status: connStatus = connStatus.IDLE) -> None:
+        """ Thread safe way of updating the connection status
+        """
+        with self._status_lock:
+            self._status = status
+
+    def update_execution_status(self, status: executionStatus = executionStatus.OK) -> None:
+        """ Thread safe way of updating the execution status
+        """
+        with self._execution_status_lock:
+            self._execution_status = status
+
     def sanitize_search_string(self, term) -> str:
         if term is not None and len(term):
-            res = sub("(_|%)", self.search_escapepattern, term)
+            res = sub("(_|%)", self.search_escapepattern + "\\1", term)
         else:
             res = term
-        return term
+        return res
+
+    def unsanitize_search_string(self, term) -> str:
+        if term is not None and len(term):
+            res = sub(self.search_escapepattern, "", term)
+        else:
+            res = term
+        return res
+
+    def escape_name(self, name):
+        if name:
+            qtchar = self.quotechar
+            name = (qtchar + "%s" + qtchar) % name
+        return name
+
+    def escape_names(self, names):
+        return [self.escape_name(name) for name in names]
+
+    def unescape_name(self, name):
+        """ Unquote a string."""
+        if name:
+            qtchar = self.quotechar
+            if name and name[0] == qtchar and name[-1] == qtchar:
+                name = name[1:-1]
+
+        return name
 
     def connect(
             self,
             username: str = "",
             password: str = "",
-            force: bool = False,
-            start_executor: bool = False) -> None:
+            force: bool = False) -> None:
         uid = username or self.username
         pwd = password or self.password
         conn_str = "DSN=" + self.dsn + ";"
@@ -99,101 +183,116 @@ class sqlConnection:
         if force or not self.conn.connected():
             try:
                 self.conn = connect(dsn = conn_str, timeout = 5)
-                self.status = connStatus.IDLE
+                self.update_status(connStatus.IDLE)
             except ConnectError as e:
                 self.logger.error("Error while connecting: %s", str(e))
                 raise ConnectError(e)
-        if start_executor:
-            self.executor = Process(
-                    target = executor_process,
-                    args=(self.child_chan, self.logger.getEffectiveLevel(),))
-            self.executor.start()
-            self.logger.info("Started executor process: %d", self.executor.pid)
-            self.parent_chan.send(cmsg("connect", conn_str, None))
-            resp = self.parent_chan.recv()
-            # How do you handle failure here?
-            if not resp.status == commandStatus.OK:
-                self.logger.error("Error atempting to connect in executor process")
-                self.executor.terminate()
-                self.executor.join()
-                raise ConnectError("Connection failure in executor")
 
-    def async_lastresponse(self) -> cmsg:
-        if self.executor and self.executor.is_alive():
-            self.logger.debug("Asking for last message, executor pid %d",
-                    self.executor.pid)
-            self.parent_chan.send(cmsg("lastresponse", None, None))
-            resp = self.parent_chan.recv()
-            # Above should never fail
-            return resp
+    def fetchmany(self, size) -> list:
+        with self._lock:
+            if self.cursor:
+                res = self.cursor.fetchmany(size)
+            else:
+                res = []
+
+        if len(res) < 1:
+            self.update_status(connStatus.IDLE)
+        with self._fetch_cv:
+            self._fetch_res.extend(res)
+            self._fetch_cv.notify()
 
-    def async_execute(self, query) -> cmsg:
-        if self.executor and self.executor.is_alive():
-            self.logger.debug("Sending query %s to pid %d",
-                    query, self.executor.pid)
-            # TODO: message should carry
-            # current catalog.  One might
-            # think that the main process
-            # connection always "follows"
-            # database changes since all
-            # main queries get executed
-            # against executor thread
-            # and main process conn only
-            # gets used for sidebar/auto
-            # completion.  But, for
-            # example the MYSQL driver
-            # if starting without a
-            # declared database will just
-            # switch to the first db
-            # when running find_columns
-            self.parent_chan.send(
-                    cmsg("execute", query, None))
-            # Will block but can be interrupted
-            res = self.parent_chan.recv()
-            self.logger.debug("Execution done")
-            self.query = query
-            # Check if catalog has changed in which case
-            # execute query locally
-            self.parent_chan.send(cmsg("currentcatalog", None, None))
-            rescat = self.parent_chan.recv()
-            if rescat.status == commandStatus.FAIL:
-                # TODO raise exception here since
-                # connection catalogs are possibly out of sync
-                # and we don't have a way of knowing
-                res = cmsg("execute", "", commandStatus.FAIL)
-            elif not rescat.payload == self.current_catalog():
-                # query changed the catalog
-                # so let's change the database locally
-                self.logger.debug("Execution changed catalog")
-                self.execute("USE " + rescat.payload)
-        else:
-            res = cmsg("execute", "", commandStatus.FAIL)
         return res
 
-    def async_fetch(self, size) -> cmsg:
-        if self.executor and self.executor.is_alive():
-            self.logger.debug("Fetching size %d from pid %d",
-                    size, self.executor.pid)
-            self.parent_chan.send(cmsg("fetch", size, None))
-            res = self.parent_chan.recv()
-            self.logger.debug("Fetching done")
-        else:
-            res = cmsg("fetch", "", commandStatus.FAIL)
+    def async_fetchall(self, size, app) -> None:
+        """ True asynchronous fetch.  Will start a fetch, that will fetch
+            *all* results (in chunks of size = size) in a background operation
+            until the result set is depleted or we signal a stop via
+            _cancel_async_event.  After thread operation is completed,
+            it asks the running event loop to redraw the app to pick up the
+            new connection status (IDLE).
+            """
+        self._cancel_async_event.clear()
+        self.update_status(connStatus.FETCHING)
+        loop = get_event_loop()
+        def _run():
+            while True:
+                res = self.fetchmany(size)
+                if len(res) < 1 or self._cancel_async_event.is_set():
+                    self.update_status(connStatus.IDLE)
+                    self._cancel_async_event.clear()
+                    # Should we try close cursor here? Problem is that
+                    # close_curor attempts to call cancel async which would
+                    # block until this thread is over
+                    break
+            loop.call_soon_threadsafe(app.invalidate)
+            return
+
+        self._fetch_thread = Thread(target = _run, daemon = True)
+        self._fetch_thread.start()
+        return
+
+    def fetch_from_cache(self, size, wait = False) -> list:
+        """ Will grab the first size elements from self._fetch_res.  Recall
+            self._fetch_res is the result cache that is built up via an async
+            fetch that grabs rows in chunks.  Here, in a threadsafe manner we
+            wait for a the asynchronous method to grab enough elements or
+            finish the fetch operation altogether, then we 'pop' from the
+            fetch result cache.
+            """
+        with self._fetch_cv:
+            if wait:
+                self._fetch_cv.wait_for(
+                        lambda: len(self._fetch_res) > size or self.status == connStatus.IDLE)
+            res = self._fetch_res[:size]
+            del self._fetch_res[:size]
         return res
 
-    def async_fetchdone(self) -> cmsg:
-        if self.executor and self.executor.is_alive():
-            self.parent_chan.send(cmsg("fetchdone", None, None))
-            res = self.parent_chan.recv()
-        else:
-            res = cmsg("fetchdone", "", commandStatus.FAIL)
-        return res
 
-    def execute(self, query, parameters = None) -> Cursor:
+    def cancel_async_fetchall(self) -> None:
+        """ Signal fetching thread to terminate operation, then wait / block
+            until thread terminates.  Also clear the fetch result cache.
+            """
+        self._cancel_async_event.set()
+        if self._fetch_thread.is_alive():
+            self._fetch_thread.join()
+        with self._fetch_cv:
+            self._fetch_res = []
+
+    def execute(self, query, parameters = None, event: Event = None) -> Cursor:
+        self.logger.debug("Execute: %s", query)
         with self._lock:
             self.cursor = self.conn.cursor()
-            self.cursor.execute(query, parameters)
-            self.query = query
+            try:
+                self._execution_err = None
+                self.update_status(connStatus.EXECUTING)
+                self.cursor.execute(query, parameters)
+                self.update_status(connStatus.IDLE)
+                self.update_execution_status(executionStatus.OK)
+                self.query = query
+            except DatabaseError as e:
+                self.update_status(connStatus.IDLE)
+                self.update_execution_status(executionStatus.FAIL)
+                self._execution_err = str(e)
+                self.logger.warning("Execution error: %s", str(e))
+            if event is not None:
+                event.set()
+        return self.cursor
+
+    def async_execute(self, query) -> Cursor:
+        """ async_ is a misnomer here.  It does execute in a new thread
+            however it will also wait for execution to complete. At this time
+            this helps us with registering KeyboardInterrupt during cyanodbc.
+            execute only; it may evolve to have more true async-like behavior.
+            """
+        self.close_cursor()
+        exec_event = Event()
+        t = Thread(
+                target = self.execute,
+                kwargs = {"query": query, "parameters": None, "event": exec_event},
+                daemon = True)
+        t.start()
+        # Will block but can be interrupted
+        exec_event.wait()
         return self.cursor
 
     def list_catalogs(self) -> list:
@@ -207,7 +306,7 @@ class sqlConnection:
                     res = self.conn.list_catalogs()
                 self.logger.debug("list_catalogs: done")
         except DatabaseError as e:
-            self.status = connStatus.ERROR
+            self.update_status(connStatus.ERROR)
             self.logger.warning("list_catalogs: %s", str(e))
 
         return res
@@ -227,7 +326,7 @@ class sqlConnection:
                     res = self.conn.list_schemas()
                 self.logger.debug("list_schemas: done")
         except DatabaseError as e:
-            self.status = connStatus.ERROR
+            self.update_status(connStatus.ERROR)
             self.logger.warning("list_schemas: %s", str(e))
 
         return res
@@ -242,7 +341,8 @@ class sqlConnection:
 
         try:
             if self.conn.connected():
-                self.logger.debug("Calling find_tables...")
+                self.logger.debug("Calling find_tables: %s, %s, %s, %s",
+                        catalog, schema, table, type)
                 with self._lock:
                     res = self.conn.find_tables(
                         catalog = catalog,
@@ -265,7 +365,8 @@ class sqlConnection:
 
         try:
             if self.conn.connected():
-                self.logger.debug("Calling find_columns...")
+                self.logger.debug("Calling find_columns: %s, %s, %s, %s",
+                        catalog, schema, table, column)
                 with self._lock:
                     res = self.conn.find_columns(
                             catalog = catalog,
@@ -278,6 +379,52 @@ class sqlConnection:
 
         return res
 
+    def find_procedures(
+            self,
+            catalog = "",
+            schema = "",
+            procedure = "") -> list:
+        res = []
+
+        try:
+            if self.conn.connected():
+                self.logger.debug("Calling find_procedures: %s, %s, %s",
+                        catalog, schema, procedure)
+                with self._lock:
+                    res = self.conn.find_procedures(
+                        catalog = catalog,
+                        schema = schema,
+                        procedure = procedure)
+                self.logger.debug("find_procedures: done")
+        except DatabaseError as e:
+            self.logger.warning("find_procedures: %s.%s.%s: %s", catalog, schema, procedure, str(e))
+
+        return res
+
+    def find_procedure_columns(
+            self,
+            catalog = "",
+            schema = "",
+            procedure = "",
+            column = "") -> list:
+        res = []
+
+        try:
+            if self.conn.connected():
+                self.logger.debug("Calling find_procedure_columns: %s, %s, %s, %s",
+                        catalog, schema, procedure, column)
+                with self._lock:
+                    res = self.conn.find_procedure_columns(
+                            catalog = catalog,
+                            schema = schema,
+                            procedure = procedure,
+                            column = column)
+                self.logger.debug("find_procedure_columns: done")
+        except DatabaseError as e:
+            self.logger.warning("find_procedure_columns: %s.%s.%s, column %s: %s", catalog, schema, procedure, column, str(e))
+
+        return res
+
     def current_catalog(self) -> str:
         if self.conn.connected():
             return self.conn.catalog_name
@@ -296,9 +443,6 @@ class sqlConnection:
         return self.conn.get_info(code)
 
     def close(self) -> None:
-        if self.executor and self.executor.is_alive():
-            self.executor.terminate()
-            self.executor.join()
         # TODO: When disconnecting
         # We likely don't want to allow any exception to
         # propagate.  Catch DatabaseError?
@@ -306,29 +450,47 @@ class sqlConnection:
             self.conn.close()
 
     def close_cursor(self) -> None:
+        self.cancel_async_fetchall()
         if self.cursor:
             self.cursor.close()
             self.cursor = None
         self.query = None
+        self.update_status(connStatus.IDLE)
 
-    def preview_query(self, table, filter_query = "", limit = -1) -> str:
-        qry = "SELECT * FROM " + table + " " + filter_query
-        if limit > 0:
-            qry = qry + " LIMIT " + str(limit)
+    def cancel(self) -> None:
+        self.cancel_async_fetchall()
+        if self.cursor:
+            self.cursor.cancel()
+        self.query = None
+        self.update_status(connStatus.IDLE)
+
+    def preview_query(
+            self,
+            name,
+            obj_type = "table",
+            filter_query = "",
+            limit = -1) -> str:
+        """ Currently we only have a generic implementation for tables and
+            views.  Otherwise (functions) we return None
+            """
+        if obj_type == "table" or obj_type == "view":
+            qry = "SELECT * FROM " + name + " " + filter_query
+            if limit > 0:
+                qry = qry + " LIMIT " + str(limit)
+        else:
+            qry = None
         return qry
 
-    def formatted_fetch(self, size, format_name = "psql"):
+    def formatted_fetch(self, size, cols, format_name = "psql"):
         while True:
-            res = self.async_fetch(size)
-            if (res.status == commandStatus.FAIL) or (not res.type == "fetch"):
-                return "Encountered a problem while fetching"
-            elif len(res.payload[1]) == 0:
+            res = self.fetch_from_cache(size)
+            if len(res) < 1 and self.status != connStatus.FETCHING:
                 break
-            else:
+            if len(res) > 0:
                 yield "\n".join(
                         formatter.format_output(
-                            res.payload[1],
-                            res.payload[0],
+                            res,
+                            cols,
                             format_name = format_name))
 
 connWrappers = {}
@@ -368,12 +530,17 @@ class MSSQL(sqlConnection):
               "'db_denydatawriter')"
 
         if catalog is None and self.current_catalog():
-            catalog = self.sanitize_search_string(self.current_catalog())
+            catalog_local = self.current_catalog()
+        else:
+            # We are going to be outright executing, versus
+            # using the ODBC API.
+            # let's make sure there is nothing escaped here
+            catalog_local = self.unsanitize_search_string(catalog)
 
-        if catalog:
+        if catalog_local:
             try:
                 self.logger.debug("Calling list_schemas...")
-                crsr = self.execute(qry.format(catalog = catalog))
+                crsr = self.execute(qry.format(catalog = catalog_local))
                 res = crsr.fetchall()
                 crsr.close()
                 self.logger.debug("Calling list_schemas: done")
@@ -385,16 +552,31 @@ class MSSQL(sqlConnection):
 
         return super().list_schemas(catalog = catalog)
 
-    def mssql_preview_query(
+    def preview_query(
             self,
-            table,
+            name,
+            obj_type = "table",
             filter_query = "",
             limit = -1) -> str:
-        qry = " * FROM " + table + " " + filter_query
-        if limit > 0:
-            qry = "SELECT TOP " + str(limit) + qry
+
+        if obj_type == "table" or obj_type == "view":
+            qry = " * FROM " + name + " " + filter_query
+            if limit > 0:
+                qry = "SELECT TOP " + str(limit) + qry
+            else:
+                qry = "SELECT" + qry
+        elif obj_type == "function":
+            # Sproc names in SQLServer come back with
+            # catalog.schema.name;INT with the trailing suffix
+            # not useful
+            name_sanitized = sub("(;\\d{0,})(\")$", "\\2", name)
+            catalog_local = sub("(.*)\\.(.*)\\.(.*)", "\\1", name)
+            qry = "SELECT definition FROM {catalog}.sys.sql_modules " \
+                  "WHERE object_id = (OBJECT_ID(N'{name}'))"
+            qry = qry.format(catalog = catalog_local, name = name_sanitized)
         else:
-            qry = "SELECT" + qry
+            qry = None
+
         return qry
 
 class PSSQL(sqlConnection):
@@ -417,6 +599,23 @@ class PSSQL(sqlConnection):
                 table = table,
                 type = type)
 
+    def find_procedures(
+            self,
+            catalog = "",
+            schema = "",
+            procedure = "") -> list:
+        """ At least the psql odbc driver I am using has an annoying habbit
+            of treating the catalog and schema fields interchangible, which
+            in turn screws up with completion"""
+
+        if not catalog in [self.current_catalog(), self.sanitize_search_string(self.current_catalog())]:
+            return []
+
+        return super().find_procedures(
+                catalog = catalog,
+                schema = schema,
+                procedure = procedure)
+
     def find_columns(
             self,
             catalog = "",
@@ -436,6 +635,29 @@ class PSSQL(sqlConnection):
                 table = table,
                 column = column)
 
+    def find_procedure_columns(
+            self,
+            catalog = "",
+            schema = "",
+            procedure = "",
+            column = "") -> list:
+        """ At least the psql odbc driver I am using has an annoying habbit
+            of treating the catalog and schema fields interchangible, which
+            in turn screws up with completion.  In addition wildcards in the column
+            field, seem to not work - but an empty string does."""
+
+        if not catalog in [self.current_catalog(), self.sanitize_search_string(self.current_catalog())]:
+            return []
+
+        if column == "%":
+            column = ""
+
+        return super().find_procedure_columns(
+                catalog = catalog,
+                schema = schema,
+                procedure = procedure,
+                column = column)
+
 class MySQL(sqlConnection):
 
     def list_schemas(self, catalog = None) -> list:
@@ -448,6 +670,15 @@ class MySQL(sqlConnection):
             we do, in-fact fall-back to find_tables"""
         return []
 
+    def current_catalog(self) -> str:
+        if self.conn.connected():
+            res = self.conn.catalog_name
+        if res == "null":
+            res = ""
+        return res
+
+class Snowflake(sqlConnection):
+
     def find_tables(
             self,
             catalog = "",
@@ -455,43 +686,14 @@ class MySQL(sqlConnection):
             table = "",
             type = "") -> list:
 
-        if catalog in ["", "null"] and schema not in ["", "null"]:
-            catalog = schema
-            schema = ""
-
+        type = type.upper()
         return super().find_tables(
                 catalog = catalog,
                 schema = schema,
                 table = table,
                 type = type)
 
-    def find_columns(
-            self,
-            catalog = "",
-            schema = "",
-            table = "",
-            column = "") -> list:
-
-        if catalog in ["", "null"] and schema not in ["", "null"]:
-            catalog = schema
-            schema = ""
-
-        return super().find_columns(
-                catalog = catalog,
-                schema = schema,
-                table = table,
-                column = column)
-
-class SQLite(sqlConnection):
-
-    def list_schemas(self, catalog = None) -> list:
-        """Easy peasy"""
-        return []
-    def list_catalogs(self) -> list:
-        """Easy peasy"""
-        return []
-
 connWrappers["MySQL"] = MySQL
 connWrappers["Microsoft SQL Server"] = MSSQL
-connWrappers["SQLite"] = SQLite
 connWrappers["PostgreSQL"] = PSSQL
+connWrappers["Snowflake"] = Snowflake
