@@ -12,7 +12,7 @@ from prompt_toolkit.layout.controls import FormattedTextControl, BufferControl,
 from prompt_toolkit.layout.dimension import Dimension
 from prompt_toolkit.buffer import Buffer
 from prompt_toolkit.document import Document
-from prompt_toolkit.filters import is_done
+from prompt_toolkit.filters import is_done, renderer_height_is_known
 from prompt_toolkit.layout.margins import ScrollbarMargin
 from prompt_toolkit.mouse_events import MouseEvent
 from prompt_toolkit.lexers import Lexer
@@ -98,16 +98,51 @@ class myDBObject:
             self.parent.collapse()
 
     def add_children(self, list_obj: List["myDBObject"]) -> None:
-        self.children = list_obj
-        for i in range(len(self.children) - 1):
-            self.children[i].next_object = self.children[i + 1]
-        self.children[len(self.children) - 1].next_object = self.next_object
-        self.next_object = self.children[0]
+        lst = list(filter(lambda x: x.name != "", list_obj))
+        if len(lst):
+            self.children = lst
+            for i in range(len(self.children) - 1):
+                self.children[i].next_object = self.children[i + 1]
+            self.children[len(self.children) - 1].next_object = self.next_object
+            self.next_object = self.children[0]
 
 class myDBColumn(myDBObject):
     def _expand_internal(self) -> None:
         return None
 
+class myDBFunction(myDBObject):
+    def _expand_internal(self) -> None:
+        cat = "%"
+        schema = "%"
+        # https://docs.microsoft.com/en-us/sql/odbc/reference/syntax/sqlprocedurecolumns-function?view=sql-server-ver15
+        # CatalogName cannot contain a string search pattern
+
+        if self.parent is not None:
+            if type(self.parent).__name__ == "myDBSchema":
+                schema = self.conn.sanitize_search_string(self.parent.name)
+            elif type(self.parent).__name__ == "myDBCatalog":
+                cat = self.parent.name
+            if self.parent.parent is not None:
+                if type(self.parent.parent).__name__ == "myDBCatalog":
+                    cat = self.parent.parent.name
+
+        res = self.conn.find_procedure_columns(
+                catalog = cat,
+                schema = schema,
+                procedure = self.conn.sanitize_search_string(self.name),
+                column = "%")
+
+        lst = [myDBColumn(
+            my_app = self.my_app,
+            conn = self.conn,
+            name = col.column,
+            otype = col.type_name,
+            parent = self,
+            level = self.level + 1) for col in res]
+
+        self.add_children(list_obj = lst)
+        return None
+
 class myDBTable(myDBObject):
     def _expand_internal(self) -> None:
         cat = "%"
@@ -137,9 +172,8 @@ class myDBTable(myDBObject):
             otype = col.type_name,
             parent = self,
             level = self.level + 1) for col in res]
-        if len(lst):
-            self.add_children(list_obj = lst)
 
+        self.add_children(list_obj = lst)
         return None
 
 class myDBSchema(myDBObject):
@@ -148,19 +182,55 @@ class myDBSchema(myDBObject):
         cat = self.conn.sanitize_search_string(self.parent.name) if self.parent is not None else "%"
         res = self.conn.find_tables(
                 catalog = cat,
-                schema = self.name,
+                schema = self.conn.sanitize_search_string(self.name),
                 table = "",
                 type = "")
-        lst = [myDBTable(
-            my_app = self.my_app,
-            conn = self.conn,
-            name = table.name,
-            otype = table.type,
-            parent = self,
-            level = self.level + 1) for table in res]
-        if len(lst):
-            self.add_children(list_obj = lst)
+        resf = self.conn.find_procedures(
+                catalog = cat,
+                schema = self.conn.sanitize_search_string(self.name),
+                procedure = "")
+        tables = []
+        views = []
+        functions = []
+        lst = []
+        for table in res:
+            if table.type.lower() == 'table':
+                tables.append(table.name)
+            if table.type.lower() == 'view':
+                views.append(table.name)
+            lst.append(myDBTable(
+                my_app = self.my_app,
+                conn = self.conn,
+                name = table.name,
+                otype = table.type.lower(),
+                parent = self,
+                level = self.level + 1))
+        for func in resf:
+            functions.append(func.name)
+            lst.append(myDBFunction(
+                my_app = self.my_app,
+                conn = self.conn,
+                name = func.name,
+                otype = "function",
+                parent = self,
+                level = self.level + 1))
 
+        self.conn.dbmetadata.extend_objects(
+                catalog = self.conn.escape_name(self.parent.name) if self.parent else "",
+                schema = self.conn.escape_name(self.name),
+                names = self.conn.escape_names(tables),
+                obj_type = "table")
+        self.conn.dbmetadata.extend_objects(
+                catalog = self.conn.escape_name(self.parent.name) if self.parent else "",
+                schema = self.conn.escape_name(self.name),
+                names = self.conn.escape_names(views),
+                obj_type = "view")
+        self.conn.dbmetadata.extend_objects(
+                catalog = self.conn.escape_name(self.parent.name) if self.parent else "",
+                schema = self.conn.escape_name(self.name),
+                names = self.conn.escape_names(functions),
+                obj_type = "function")
+        self.add_children(list_obj = lst)
         return None
 
 class myDBCatalog(myDBObject):
@@ -169,36 +239,56 @@ class myDBCatalog(myDBObject):
         schemas = self.conn.list_schemas(
                 catalog = self.conn.sanitize_search_string(self.name))
 
-        if len(schemas) < 1:
+        if len(schemas) < 1 or all([s == "" for s in schemas]):
             res = self.conn.find_tables(
                     catalog = self.conn.sanitize_search_string(self.name),
                     schema = "",
                     table = "",
                     type = "")
-            for r in res:
-                if (r.schema not in schemas and r.schema != ""):
-                    schemas.append(r.schema)
-        if len(schemas):
+            schemas = [r.schema for r in res]
+
+        self.conn.dbmetadata.extend_schemas(
+                catalog = self.conn.escape_name(self.name),
+                names = self.conn.escape_names(schemas))
+
+        if not all([s == "" for s in schemas]):
+            # Schemas were found either having called list_schemas
+            # or via the find_tables call
             lst = [myDBSchema(
                 my_app = self.my_app,
                 conn = self.conn,
                 name = schema,
-                otype = "Schema",
+                otype = "schema",
                 parent = self,
-                level = self.level + 1) for schema in schemas]
-        else:
+                level = self.level + 1) for schema in sorted(set(schemas))]
+        elif len(schemas):
             # No schemas found; but if there are tables then these are direct
             # descendents, i.e. MySQL
-            lst = [myDBTable(
-                my_app = self.my_app,
-                conn = self.conn,
-                name = table.name,
-                otype = table.type,
-                parent = self,
-                level = self.level + 1) for table in res]
-        if len(lst):
-            self.add_children(list_obj = lst)
+            tables = []
+            views = []
+            lst = []
+            for table in res:
+                if table.type.lower() == 'table':
+                    tables.append(table.name)
+                if table.type.lower() == 'view':
+                    views.append(table.name)
+                lst.append(myDBTable(
+                    my_app = self.my_app,
+                    conn = self.conn,
+                    name = table.name,
+                    otype = table.type.lower(),
+                    parent = self,
+                    level = self.level + 1))
+            self.conn.dbmetadata.extend_objects(
+                    catalog = self.conn.escape_name(self.name),
+                    schema = "", names = self.conn.escape_names(tables),
+                    obj_type = "table")
+            self.conn.dbmetadata.extend_objects(
+                    catalog = self.conn.escape_name(self.name),
+                    schema = "", names = self.conn.escape_names(views),
+                    obj_type = "view")
 
+        self.add_children(list_obj = lst)
         return None
 
 
@@ -216,37 +306,51 @@ class myDBConn(myDBObject):
                     my_app = self.my_app,
                     conn = self.conn,
                     name = row,
-                    otype = "Catalog",
+                    otype = "catalog",
                     parent = self,
                     level = self.level + 1) for row in rows]
+                self.conn.dbmetadata.extend_catalogs(
+                        self.conn.escape_names(rows))
         else:
-            schemas = []
             res = self.conn.find_tables(
                     catalog = "%",
                     schema = "",
                     table = "",
                     type = "")
-            for r in res:
-                if (r.schema not in schemas and r.schema != ""):
-                    schemas.append(r.schema)
-            if len(schemas):
+            schemas = [r.schema for r in res]
+            self.conn.dbmetadata.extend_schemas(catalog = "",
+                    names = self.conn.escape_names(schemas))
+            if not all([s == "" for s in schemas]):
                 lst = [myDBSchema(
                     my_app = self.my_app,
                     conn = self.conn,
                     name = schema,
-                    otype = "Schema",
+                    otype = "schema",
                     parent = self,
-                    level = self.level + 1) for schema in schemas]
-            else:
-                lst = [myDBTable(
+                    level = self.level + 1) for schema in sorted(set(schemas))]
+            elif len(schemas):
+                tables = []
+                views = []
+                lst = []
+                for table in res:
+                    if table.type.lower() == 'table':
+                        tables.append(table.name)
+                    if table.type.lower() == 'view':
+                        views.append(table.name)
+                    lst.append(myDBTable(
                     my_app = self.my_app,
                     conn = self.conn,
                     name = table.name,
-                    otype = table.type,
+                    otype = table.type.lower(),
                     parent = self,
-                    level = self.level + 1) for table in res]
-        if len(lst):
-            self.add_children(list_obj = lst)
+                    level = self.level + 1))
+                self.conn.dbmetadata.extend_objects(catalog = "",
+                        schema = "", names = self.conn.escape_names(tables),
+                        obj_type = "table")
+                self.conn.dbmetadata.extend_objects(catalog = "",
+                        schema = "", names = self.conn.escape_names(views),
+                        obj_type = "view")
+        self.add_children(list_obj = lst)
         return None
 
 def sql_sidebar_help(my_app: "sqlApp"):
@@ -365,7 +469,7 @@ def show_sidebar_button_info(my_app: "sqlApp") -> Container:
             & Condition(
                 lambda: not my_app.show_exit_confirmation
             )
-#            & renderer_height_is_known
+            & renderer_height_is_known
             )
 
 def sql_sidebar(my_app: "sqlApp") -> Window:
